/* pyramid.c (Combinatorial Pyramids project)
 *  Copyright (C) 2003 Luc Brun
 *  email: luc.brun@univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <float.h>
#define __USE_MOVE_TYPE__
#include "tab_int.h"
#include "tab_dart.h"
#include "tab_gen.h"
#include "tab_mark_dart.h"
#include "level_set.h"
#include "dart_list.h"
#include "image.h"
#include "objet.h"

#include "pyramid.h"

#include "move_type.h"

#define ABS(a) ((a)<0?(-(a)):(a))


#define Sigma          SELF(sigma)
#define Sigma_base     SELF(sigma_base)
#define Background     SELF(background)
#define Mu             SELF(mu)
#define Mu_1           SELF(mu_1)
#define Node_values    SELF(node_values)
#define Levels         SELF(levels)
#define Vertex_set     SELF(vertex_set)
#define Dimx           SELF(dimx)
#define Dimy           SELF(dimy)
#define Index_or       SELF(index_or)
#define Dart_emb       SELF(dart_emb)
#define Vertex_value   SELF(vertex_value)
#define Dart_value     SELF(dart_value)

struct pyramid
{
  tab_dart sigma;
  tab_dart sigma_base;
  tab_dart mu;
  tab_int mu_1;
  tab_int index_or;
  tab_dart dart_emb;
  int dimx;
  int dimy;
  unsigned long background;
  level_set levels;
  dart_list vertex_set;
  vertex_data* vertex_value;
  dart_data* dart_value;
};
/* Private methods. */
static int  find_compress_priv(tab_int,vertex);
static dart sigma_1_priv(pyramid,dart);
static void do_removals_priv(pyramid,dart_list,PFmerge_edge_value);
static void remove_double_edges_priv(pyramid,PFmerge_edge_value);
static void remove_empty_self_loops_priv(pyramid,dart_list);

static void draw_internal_boundary_priv(pyramid,
					PFdraw_point,
					tab_int,
					int,
					tab_dart,
					dart);
static void draw_point_priv(pyramid,PFdraw_point,dart);
static dart next_boundary_dart_priv(pyramid,tab_int,int,dart);
static void do_contractions_priv(pyramid,tab_mark_dart,dart_list);
static void update_non_survivor_priv(pyramid,
				     tab_mark_dart,
				     tab_int,
				     tab_mark_dart,
				     PFequivalent ,
				     PFparent_child);
static int update_survivor_priv(pyramid,
				    tab_mark_dart,
				    tab_mark_dart,
				    PFequivalent ,
				    PFvalue);


static tab_int init_father_priv(pyramid,tab_int);
static void update_vertex_set_priv(pyramid,tab_int,PFmerge);
static void compute_pixel_value_priv(pyramid,dart,dart,tab_int,int *);
static void init_pyramid_priv(pyramid,int,int);

static void load_update_sigma_priv(pyramid,tab_int);
static dart load_compute_sig_succ_priv(pyramid,tab_int,dart);
static void load_set_vertice_priv(pyramid,tab_int);
static void init_vertex_set_priv(pyramid);
static int  print_sigma_orbit_priv(pyramid,dart);
static void update_orientation_priv(pyramid,dart,dart);
static  dart_list get_open_parenthesis_priv(pyramid,vertex,tab_mark_dart);
static tab_mark_dart get_inclusions_priv(pyramid,vertex,dart_list,tab_mark_dart);

extern pyramid new_pyramid(int dimx,int dimy, vertex_data *v_data, dart_data *b_data)
{
  dart b;
  int nb_edge;
  INSTANCIER(pyramid);
  
  assert(dimx&&dimy);
  
  nb_edge = dimx*(dimy+1)+(dimx+1)*dimy;
  Levels=new_level_set(nb_edge);
  init_pyramid_priv(self,dimx,dimy);
  init_vertex_set_priv(self);
  
  /* copy of the permutation sigma_base into sigma*/
  for(b=1;b<=nb_edge;b++)
    {
      tab_dart_put_value(Dart_emb,b,b);
      tab_dart_put_value(Dart_emb,-b,-b);
      tab_dart_put_value(Sigma,b,tab_dart_give_value(Sigma_base,b));
      tab_dart_put_value(Sigma,-b,tab_dart_give_value(Sigma_base,-b));
    }

  Vertex_value=v_data;
  Dart_value=b_data;

  pyramid_init_node_value(self,Vertex_value->init);
  pyramid_init_edge_value(self,Dart_value->init);

  return self;
}

extern pyramid pyramid_init_from_file(char * name)
{
  int dimx,dimy;
  tab_int levels;
  image img;
  
  INSTANCIER(pyramid);
  img=FAIRE_image();
  assert(image_charger(img,name)==0);
  assert(image_give_dim(img)==3);
  dimx=image_give_largeur(img)-1;
  dimy=image_give_hauteur(img)-1;
  
  levels=FAIRE_tab_int();
  Levels=level_set_load(img,levels);
  init_pyramid_priv(self,dimx,dimy);
  load_update_sigma_priv(self,levels);
  load_set_vertice_priv(self,levels);
  
  DEFAIRE_tab_int(levels);
  DEFAIRE_image(img);

  return self;
}
extern void destroy_pyramid(pyramid self,
                            PFfree free_obj)
{  
  DEFAIRE_tab_dart(Sigma);
  DEFAIRE_tab_dart(Sigma_base);
  DEFAIRE_tab_int(Mu_1);
  DEFAIRE_tab_dart(Mu);  
  DEFAIRE_dart_list(Vertex_set);
  destroy_pyramid_value(Vertex_value->data);
  destroy_level_set(Levels);
  destroy_border_value(Dart_value->data);
  
  free(self);
}


extern void pyramid_init_node_value(pyramid self, 
                                    PFinit_node_value init)
{
  int mu;

  dart_list_first(Vertex_set);
  do
    {
      mu=dart_list_give_current(Vertex_set);
      Vertex_value->init(Vertex_value->data,self,mu);
    }
  while(dart_list_next(Vertex_set));
}

extern void pyramid_init_edge_value(pyramid self, 
                                    PFinit_edge_value init)
{
  dart b;
  float val;

  level_set_first(Levels,
		  level_set_give_max_level(Levels));
  do
    { 
      b= level_set_current(Levels);
      val=init(Dart_value->data,self,b);
      border_value_put_value(Dart_value->data,b,val);
      border_value_put_cardinal(Dart_value->data,b,1);
      val=init(Dart_value->data,self,-b);
      border_value_put_value(Dart_value->data,-b,val);
      border_value_put_cardinal(Dart_value->data,-b,1);
    }
  while(level_set_next(Levels));
}

extern vertex pyramid_give_vertex(pyramid self,dart b)
{
  return tab_dart_give_value(Mu,b);
}
extern dart_list pyramid_give_vertex_set(pyramid self)
{
  return dart_list_copy(Vertex_set);
}

extern int pyramid_give_nb_regions(pyramid self)
{
  return dart_list_give_cardinal(Vertex_set);
}
extern dart pyramid_give_sigma(pyramid self,dart b)
{
  assert(self);
  return tab_dart_give_value(Sigma,b);
}
extern vertex pyramid_give_background(pyramid self)
{
  assert(self);
  return Background;
}
extern void pyramid_give_undifferentiated_neighborhood(pyramid self,
                                                       vertex v,
                                                       dart_list neighbors)
{
  dart b,b_first;
  vertex mu;
  tab_mark_dart encountered;
  
  encountered=FAIRE_tab_mark_dart();
  b=b_first=tab_int_give_value(Mu_1,v);
  do
    {
      mu=tab_dart_give_value(Mu,-b);
      if((mu != v) && !tab_mark_dart_is_marked(encountered,mu))
	{
	  dart_list_insert(neighbors,mu);
	  tab_mark_dart_mark_dart(encountered,mu);
	}
      b=tab_dart_give_value(Sigma,b);
    }
  while(b!=b_first);
  DEFAIRE_tab_mark_dart(encountered);
}
extern void pyramid_give_neighbors(pyramid self,
                                   vertex v,
                                   dart_list neighbors,
                                   dart_list included)
{
  dart b,b_first;
  vertex mu;
  tab_mark_dart start_loop,neighbor_mark,included_mark;
  
  neighbor_mark=FAIRE_tab_mark_dart();
  included_mark=FAIRE_tab_mark_dart();
  
  start_loop=get_inclusions_priv(self,v,included,included_mark);

  if(neighbors == NULL_DART_LIST)
    {
      DEFAIRE_tab_mark_dart(start_loop);
      DEFAIRE_tab_mark_dart(neighbor_mark);
      DEFAIRE_tab_mark_dart(included_mark);
      return;
    }      

  b=b_first=tab_int_give_value(Mu_1,v);
  do
    {
      mu=tab_dart_give_value(Mu,-b);
      if((mu!=v)&&(!tab_mark_dart_is_marked(included_mark,mu)))
	{
	  if(!tab_mark_dart_is_marked(neighbor_mark,mu))
	    dart_list_insert(neighbors,mu);
	}
      if(tab_mark_dart_is_marked(start_loop,b))
	b=-b;
      else
	b=tab_dart_give_value(Sigma,b);
    }
  while(b != b_first);

  DEFAIRE_tab_mark_dart(start_loop);
  DEFAIRE_tab_mark_dart(neighbor_mark);
  DEFAIRE_tab_mark_dart(included_mark);
}

extern void pyramid_sequential_reduce(pyramid self, 
                                      sequential_seg *seq_seg)
{
  vertex mu_b,mu_mb;
  vertex f_mu,f_mu_mb;
  dart b;
  dart_list empty_self_loops;
  int card_cont_kernel=0;
  PFmerge merge_node;
  PFmerge_edge_value merge_border;

  if (seq_seg->merge==NULL_PFMERGE)
    merge_node=Vertex_value->merge;
  else
    merge_node=seq_seg->merge;
    
  if (seq_seg->merge_dart==NULL_PFMERGE_DART)
    merge_border=Dart_value->merge;
  else
    merge_border=seq_seg->merge_dart;
    
  
  tab_int father;
  tab_mark_dart cont_kernel=FAIRE_tab_mark_dart();
  
  father=init_father_priv(self,NULL_TAB_INT);
  
  
  level_set_first(Levels,
		  level_set_give_max_level(Levels));
  do
    { 
      b= level_set_current(Levels);
      mu_b=tab_dart_give_value(Mu,b);
      mu_mb=tab_dart_give_value(Mu,-b);
      if(mu_b==Background || mu_mb==Background)	/* We do not contract
						   the background */
	continue;
      
      f_mu=find_compress_priv(father,mu_b);
      f_mu_mb=find_compress_priv(father,mu_mb);
      
      if (f_mu == f_mu_mb)			/* mu and mu_mb belong:
						   to the same tree*/
	continue;
      if(seq_seg->equiv(self,
		        level_set_give_max_level(Levels)-1,
		        mu_b,mu_mb,b))
	{
	  card_cont_kernel++;
	  tab_mark_dart_mark_dart(cont_kernel,b);
	  tab_mark_dart_mark_dart(cont_kernel,-b);
	  if(f_mu<f_mu_mb)
	    tab_int_put_value(father,f_mu_mb,f_mu);
	  else
	    tab_int_put_value(father,f_mu,f_mu_mb);
	}
    }
  while(level_set_next(Levels));
  
  if(card_cont_kernel == 0)			/* we done nothing */
    {
      DEFAIRE_tab_int(father);
      return;
    }
  level_set_new_level(Levels);
  empty_self_loops=FAIRE_dart_list();
  
  do_contractions_priv(self,cont_kernel,empty_self_loops);
  do_removals_priv(self,empty_self_loops,merge_border);
  DEFAIRE_dart_list(empty_self_loops);
  /* We update the vertex of each dart*/
  update_vertex_set_priv(self,father,merge_node);
  
  /*level_set_debug(Levels);*/
  DEFAIRE_tab_int(father);
}

extern void pyramid_parallel_reduce(pyramid self, parallel_seg *par_seg)
{
  tab_mark_dart p,q,cont_kernel;
  tab_int father=NULL_TAB_INT;
  dart_list empty_self_loops;
  int nb_vertice,nb_vertice2;
  int iter=0;
  PFmerge merge_node;
  PFmerge_edge_value merge_border;

  if (par_seg->merge==NULL_PFMERGE)
    merge_node=Vertex_value->merge;
  else
    merge_node=par_seg->merge;
    
  if (par_seg->merge_dart==NULL_PFMERGE_DART)
    merge_border=Dart_value->merge;
  else
    merge_border=par_seg->merge_dart;

  p=FAIRE_tab_mark_dart();
  q=FAIRE_tab_mark_dart();
  empty_self_loops=FAIRE_dart_list();

  cont_kernel=FAIRE_tab_mark_dart();

  nb_vertice2=dart_list_give_cardinal(Vertex_set)-1;
  printf("Nb vertices Av: %i\n", nb_vertice2);
  do
  {
    iter++;
    father=init_father_priv(self,father);
    nb_vertice=nb_vertice2;
    nb_vertice2=0;

    nb_vertice2=update_survivor_priv(self,p,q,par_seg->equiv,par_seg->value);
    update_non_survivor_priv(self,p,father,cont_kernel,par_seg->equiv,par_seg->interest);

    if(nb_vertice == nb_vertice2)
      break;
    level_set_new_level(Levels);
    do_contractions_priv(self,cont_kernel,empty_self_loops);
    do_removals_priv(self,empty_self_loops,merge_border);
    update_vertex_set_priv(self,father,merge_node);
    fprintf(stderr,"Iteration : %d\n",iter);
    tab_mark_dart_reinitialize(cont_kernel);
    tab_mark_dart_reinitialize(p);
    tab_mark_dart_reinitialize(q);
    dart_list_reinitialize(empty_self_loops);
  }
  while(1);
  
  printf("Nb vertices Ap: %i\n", nb_vertice2);
  
  DEFAIRE_tab_mark_dart(p);
  DEFAIRE_tab_mark_dart(q);
  DEFAIRE_tab_mark_dart(cont_kernel);
  DEFAIRE_tab_int(father);
  DEFAIRE_dart_list(empty_self_loops);
}
extern void pyramid_draw_internal_boundaries(pyramid self,
                                             draw_mode mode,
                                             PFnew_segment new_dart,
                                             PFdraw_point draw)
{
  vertex mu;
  tab_int levels;

  levels=level_set_give_all_levels(Levels);

  dart_list_first(Vertex_set);
  do
  {
    mu=dart_list_give_current(Vertex_set);
    if(mu == Background)
      continue;
    pyramid_draw_vertex(self,mu,levels,mode,new_dart,draw);
  }
  while(dart_list_next(Vertex_set));

  DEFAIRE_tab_int(levels);
}
extern void pyramid_draw_vertex(pyramid self,
                                vertex v,
                                tab_int levels,
                                draw_mode mode,
                                PFnew_segment new_dart,
                                PFdraw_point draw)
{
  tab_dart sigma=Sigma;
  dart b,b_first;
  int current_level=level_set_give_max_level(Levels);
  vertex v_mb;

  b_first=b=tab_int_give_value(Mu_1,v);
  do
  {
    if(mode != pyr_draw_all)
    {
      v_mb=tab_dart_give_value(Mu,-b);
      if(v == v_mb)
        continue; 
    }
    if((mode == pyr_draw_single) && (b<0))
      continue;
    if(new_dart != NULL_PFNEW_SEGMENT && !new_dart(self,b))
      continue;

    draw_internal_boundary_priv(self,
    				draw,
    				levels,current_level,
    				sigma,b);
  }
  while((b=tab_dart_give_value(sigma,b)) != b_first);
}

extern tab_int pyramid_give_levels(pyramid self)
{
  return level_set_give_all_levels(Levels);
}
extern void pyramid_save(pyramid self,char* name)
{
  FILE* id=fopen(name,"w");
  pyramid_to_stream(self,id);
  fclose(id);
}
extern void pyramid_to_stream(pyramid self,FILE* id)
{
  int value[3];
  int i,j;
  dart right,down;
  tab_int levels=level_set_give_all_levels(Levels);
  image img=FAIRE_image();
  image_initialize(img,
		   3,
		   Dimx+1,
		   Dimy+1);
  right=1;
  down=Dimx*(Dimy+1)+1;
  image_debut(img);
  for(i=0;i<Dimy;i++)
    {
      for(j=0;j<Dimx;j++)
	{
	  compute_pixel_value_priv(self,right,down,levels,value);
	  image_ecrire_pixel(img,value);
	  assert(image_pixel_suivant(img));
	  right++;
	  down++;
	}
      compute_pixel_value_priv(self,0,down,levels,value);
      image_ecrire_pixel(img,value);
      assert(image_pixel_suivant(img));
      down++;
    }
  for(j=0;j<Dimx;j++)
    {
      compute_pixel_value_priv(self,right,0,levels,value);
      image_ecrire_pixel(img,value);
      assert(image_pixel_suivant(img));
      right++;
    }
  image_to_stream(img,id);
  DEFAIRE_image(img);
  DEFAIRE_tab_int(levels);
}
static void compute_pixel_value_priv(pyramid self,
				     dart right,
				     dart down,
				     tab_int levels,
				     int *value)
{
  int level;

  value[1]=0;
  if(right != 0)
    {
      level=tab_int_give_value(levels,right);
      value[0]=level&0xff;
      value[1]=(level>>8)&0x0f;
    }
  if(down!=0)
    {
      level=tab_int_give_value(levels,down);
      value[1]=(value[1])|((0x0f&level)<<4);
      value[2]=level>>4;
    }
}
void pyramid_debug(pyramid self)
{
  int mu;
  dart b,b_first;
  int nb_dart;

  level_set_first(Levels,
		  level_set_give_max_level(Levels));

  do
    {
      b=level_set_current(Levels);      
      fprintf(stderr,"Sigma[%d]=%d\n",b,tab_dart_give_value(Sigma,b));
      fprintf(stderr,"Sigma[%d]=%d\n",-b,tab_dart_give_value(Sigma,-b));
    }
  while(level_set_next(Levels));
  dart_list_first(Vertex_set);
  do
    {
      mu=dart_list_give_current(Vertex_set);
      b_first=b=tab_int_give_value(Mu_1,mu);

      fprintf(stderr,"\n%d: (",mu);
      nb_dart=print_sigma_orbit_priv(self,b_first);
      fprintf(stderr,") :%d\n",nb_dart);
    }
  while(dart_list_next(Vertex_set));
  fprintf(stderr,"\n");
}
static int print_sigma_orbit_priv(pyramid self,dart b_first)
{
  dart b=b_first;
  int  nb_dart=0;
  int or_f=0;
  move_type first_move,last_move,move;
  dart phi2_b;

  phi2_b=tab_dart_give_value(Sigma,-b);
  phi2_b=tab_dart_give_value(Sigma,-phi2_b);

  first_move=give_move(self,
			    tab_dart_give_value(Dart_emb,b));
  if(b>0)
    or_f=tab_int_give_value(Index_or,b);
  else
    or_f=-tab_int_give_value(Index_or,-b);

  last_move=REVERSE_SENSE(give_move(self,
					 tab_dart_give_value(Dart_emb,-b)));
  fprintf(stderr,"%d,",b);
  nb_dart=1;
  b=tab_dart_give_value(Sigma,b);
  while(b != b_first)
    {
      fprintf(stderr,"%d,",b);

      move=give_move(self,
			  tab_dart_give_value(Dart_emb,b));
      or_f+=move_type_compute_step_clock(last_move,move);
      if(b>0)
	or_f+=tab_int_give_value(Index_or,b);
      else
	or_f-=tab_int_give_value(Index_or,-b);
      last_move=REVERSE_SENSE(give_move(self,
					     tab_dart_give_value(Dart_emb,-b)));
            
      b=tab_dart_give_value(Sigma,b);
      nb_dart++;
    }
  while(b != b_first);
  or_f+=move_type_compute_step_clock(last_move,first_move);
  fprintf(stderr,"or=%d",or_f);
  return nb_dart;
}
static int find_compress_priv(tab_int father,vertex n)
{
  vertex f=tab_int_give_value(father,n);
  
  if(f!=n)
    {
      f=find_compress_priv(father,f);
      tab_int_put_value(father,n,f);
    }
  return f;
}
static dart sigma_1_priv(pyramid self,dart b_first)
{
  dart b_prec;
  dart b=b_first;
  
  do
    {
      b_prec=b;
      b=tab_dart_give_value(Sigma,-b);
    }
  while(b != b_first);
  
  return -b_prec;
}
    
    
  
  
static void do_removals_priv(pyramid self,
			     dart_list empty_self_loops,
			     PFmerge_edge_value merge_border)
{

  level_set_new_level(Levels);
  /* We first remove empty self loops */
  remove_empty_self_loops_priv(self,empty_self_loops);
  
  /* Then we remove double edges */
  level_set_new_level(Levels);
  remove_double_edges_priv(self, merge_border);
}
static void   remove_empty_self_loops_priv(pyramid self,
					   dart_list empty_self_loops) 
{
  int front;
  vertex mu, mu_mb;

  dart b,b_succ;
  if(!dart_list_is_empty(empty_self_loops))
  {
    dart_list_first(empty_self_loops);
    do
    {
      b_succ=-dart_list_give_current(empty_self_loops);
      do
      {
        b=b_succ;

        front=border_value_give_cardinal(Dart_value->data, b);
        mu=pyramid_give_vertex(self,b);
        mu_mb=pyramid_give_vertex(self,-b);
        pyramid_value_add_perimetre(Vertex_value->data,mu,-front);
        pyramid_value_add_perimetre(Vertex_value->data,mu_mb,-front);

        level_set_suppress_edge(Levels,ABS(b));
        b_succ=tab_dart_give_value(Sigma,b);
      }
      while(tab_dart_give_value(Sigma,-b_succ)==-b);
      tab_dart_put_value(Sigma,sigma_1_priv(self,-b),b_succ);
    }
    while(dart_list_next(empty_self_loops));
  }
}
static void   remove_double_edges_priv(pyramid self, PFmerge_edge_value merge_border)
{
  dart b,bp;
  dart sig_b,phi_b,sig2;
  dart edge;

  dart_list double_edges_list=FAIRE_dart_list();
  tab_mark_dart double_edges_mark=FAIRE_tab_mark_dart();

  level_set_first(Levels,
		  level_set_give_max_level(Levels));
  do
    {
      edge=level_set_current(Levels);
      if(tab_mark_dart_is_marked(double_edges_mark,edge))
	continue;
      for(b=-edge;b<=edge;b+=2*edge)
	{
	  sig_b=tab_dart_give_value(Sigma,b);	/* sigma(b) */
	  phi_b=tab_dart_give_value(Sigma,-sig_b);	/* phi(sigma(b)) */
	  sig2=tab_dart_give_value(Sigma,sig_b);	/* sigma^2(b) */
	  bp=sigma_1_priv(self,-sig_b);

	  while((phi_b == -b)&&(sig2==-bp)&&(sig_b!=b))
	    {
	      dart_list_insert(double_edges_list,ABS(sig_b));
	      tab_mark_dart_mark_dart(double_edges_mark,ABS(sig_b));
	      tab_dart_put_value(Sigma,bp,-b);

	      update_orientation_priv(self,b,sig_b);
	      merge_border(Dart_value->data,self,edge,b,sig_b);

	      sig_b= sig2;
	      bp=sigma_1_priv(self,-sig_b);
	      sig2=tab_dart_give_value(Sigma,sig2); 
	      tab_dart_put_value(Sigma,b,sig_b);

	      phi_b=tab_dart_give_value(Sigma,-sig_b);
	    }
	}

    }
  while(level_set_next(Levels));

  if(!dart_list_is_empty(double_edges_list))
    {
      dart_list_first(double_edges_list);
      do
	{
	  b=dart_list_give_current(double_edges_list);
	  level_set_suppress_edge(Levels,b);
	}
      while(dart_list_next(double_edges_list));
    }
  DEFAIRE_dart_list(double_edges_list);
  DEFAIRE_tab_mark_dart(double_edges_mark);

}
static void draw_internal_boundary_priv(pyramid self,
					PFdraw_point draw,
					tab_int levels,
					int current_level,
					tab_dart sigma,
					dart b)
{
  dart sig_b=tab_dart_give_value(sigma,b);
  do
    {
      draw_point_priv(self,draw,b);
      b=next_boundary_dart_priv(self,levels,current_level,b);
    }
  while(b != sig_b);
}
static void draw_point_priv(pyramid self,
			    PFdraw_point draw,
			    dart b)
{
  dart num_pixel,b_curr;
  Point P;
  int pixel;

  b_curr=b;
  num_pixel=ABS(b);
  do
    {
      pixel=ABS(b_curr);
      if (num_pixel>pixel)
	num_pixel = pixel;
      b_curr=tab_dart_give_value(Sigma_base,b_curr);
    }
  while(b_curr != b);
  num_pixel--;
  COORDX(P)=num_pixel%Dimx;
  COORDY(P)=num_pixel/Dimx;
  
  draw(self,b,&P);
}
  
static dart next_boundary_dart_priv(pyramid self,
				    tab_int levels,
				    int current_level,
				    dart b)
{
  int level_b;
  b=-b;
  do
    {
      b=tab_dart_give_value(Sigma_base,-b);
      level_b=tab_int_give_value(levels,ABS(b));
    }
  while((level_b%3 != 0)&&(level_b < current_level));

  return b;
}
static int update_survivor_priv(pyramid self,
                                tab_mark_dart p,
                                tab_mark_dart q,
                                PFequivalent lambda,
                                PFvalue value)
{
   static double my_value(vertex mu)
   {
   	   if (value == NULL_PFVALUE)
   		   return (double)mu;
   	   return value(self,mu);
   }
   static bouleen my_lambda(vertex v1,vertex v2,dart b)
   {
   	   if(lambda==NULL_PFEQUIVALENT)
   		   return vrai;
   	   return lambda(self,
   			 level_set_give_max_level(Levels)-1,
   			 v1,
   			 v2,
			 b);
   }

  vertex mu,mu_mb;
  dart b,b_first;
  double min_val;
  bouleen min;
  int nb_survivor=0, nb_val=0, nb_vertex=dart_list_give_cardinal(Vertex_set)-1;

  do 
  {
    dart_list_first(Vertex_set);
    do
    {
      mu=dart_list_give_current(Vertex_set);
      if(mu == Background)
    	continue;
      if(tab_mark_dart_is_marked(p,mu))
    	continue;
      if(tab_mark_dart_is_marked(q,mu))
    	continue;

      min = vrai;
      b=b_first=tab_int_give_value(Mu_1,mu);
      min_val=my_value(mu);
      do
      {
    	mu_mb=tab_dart_give_value(Mu,-b);
    	if(mu_mb == Background)
    	  continue;
    	if(mu == mu_mb)
    	  continue;
    	if(tab_mark_dart_is_marked(p,mu_mb))
    	  continue;
    	if(tab_mark_dart_is_marked(q,mu_mb))
    	  continue;
        if(!my_lambda(mu,mu_mb,b))
          continue;
        if(min_val > my_value(mu_mb))
    	{
    	  min=faux;
    	  break;
    	}
      }
      while((b=tab_dart_give_value(Sigma,b)) != b_first);
      
      if(min)
      {
    	tab_mark_dart_mark_dart(p,mu);
    	nb_survivor++;
    	
    	b=b_first=tab_int_give_value(Mu_1,mu);
	

    	do
    	{
    	  mu_mb=tab_dart_give_value(Mu,-b);
    	  if(mu_mb == Background)
    	    continue;
    	  if(mu == mu_mb)
    	    continue;
    	  if(tab_mark_dart_is_marked(q,mu_mb))
    	    continue;
          if(!my_lambda(mu,mu_mb,b))
            continue;
    	  tab_mark_dart_mark_dart(q,mu_mb);
    	  nb_val++;
    	}
    	while((b=tab_dart_give_value(Sigma,b)) != b_first);
      }
    }
    while(dart_list_next(Vertex_set));
  }
  while((nb_survivor+nb_val)!=nb_vertex);
     
  return nb_survivor;
}

static void update_non_survivor_priv(pyramid self,
                                     tab_mark_dart p,
                                     tab_int father,
                                     tab_mark_dart cont_kernel,
                                     PFequivalent lambda,
                                     PFparent_child interest)
{
  vertex mu,mu_mb;
  dart b,b_first,b_cont;
  double curr_int,min_int=DBL_MAX; 


  dart_list_first(Vertex_set);
  do
  {
    mu=dart_list_give_current(Vertex_set);
    if(mu==Background)
      continue;
    if(tab_mark_dart_is_marked(p,mu))
      continue;
      
    b=b_cont=b_first=tab_int_give_value(Mu_1,mu);
    min_int=DBL_MAX; 
    b_cont=0;
    do
      {
	mu_mb=tab_dart_give_value(Mu,-b);
	if(mu==mu_mb)
	  continue;
	if(mu_mb==Background)
	  continue;
	if(!tab_mark_dart_is_marked(p,mu_mb))
	  continue;
	if((lambda != NULL_PFEQUIVALENT)
	   &&
	   !lambda(self,
		   level_set_give_max_level(Levels),
		   mu,
		   mu_mb,
		   b))
	  continue;
	/* If no interest operator is specified we take */
	/* the first surviving neighbor. */
	if(interest == NULL_PFPARENT_CHILD)
	  break;
	curr_int=interest(self,
			  level_set_give_max_level(Levels),
			  mu,mu_mb,b);
	if(curr_int <= min_int)
	  {
	    b_cont = b;
	    min_int=curr_int;
	  }       
      }
    while((b=tab_dart_give_value(Sigma,b))
	  !=
	  b_first);
    
    assert(b_cont!= 0);
    mu_mb=tab_dart_give_value(Mu,-b_cont);
    tab_mark_dart_mark_dart(cont_kernel,b_cont);
    tab_mark_dart_mark_dart(cont_kernel,-b_cont);
    tab_int_put_value(father,mu,mu_mb);
  }
  while(dart_list_next(Vertex_set));
}
static void do_contractions_priv(pyramid self,
                                 tab_mark_dart cont_kernel,
                                 dart_list empty_self_loops)
{
  dart b,b_succ;
  int edge, front;
  vertex mu, mu_mb;
  dart_list contracted =FAIRE_dart_list();
  
  level_set_first(Levels,
		  level_set_give_max_level(Levels));
  do
  {
    edge=level_set_current(Levels);
    if(tab_mark_dart_is_marked(cont_kernel,edge))
    {
      front=border_value_give_cardinal(Dart_value->data, edge);
      mu=pyramid_give_vertex(self,edge);
      mu_mb=pyramid_give_vertex(self,-edge);
      pyramid_value_add_perimetre(Vertex_value->data,mu,-front);
      pyramid_value_add_perimetre(Vertex_value->data,mu_mb,-front);
      dart_list_insert(contracted,edge);
      continue;
    }

    for(b=-edge;b<=edge;b+=2*edge)	      /* we consider edge
  						 and -edge */
    {
      /* we compute the sigma successor */
      b_succ=tab_dart_give_value(Sigma,b);
      while(tab_mark_dart_is_marked(cont_kernel,b_succ))
  	b_succ=tab_dart_give_value(Sigma,-b_succ);
      tab_dart_put_value(Sigma,b,b_succ);
      /* check if it is an empty self-loop */
      if(b_succ== -b)		
      {
  	dart_list_insert(empty_self_loops,b);
  	continue;
      }
    }
  }
  while(level_set_next(Levels));
  
  if(dart_list_is_empty(contracted))
    {
      DEFAIRE_dart_list(contracted);
      return;
    }
  dart_list_first(contracted);
  do
    {
      edge=dart_list_give_current(contracted);
      level_set_suppress_edge(Levels,edge);
    }
  while(dart_list_next(contracted));
  DEFAIRE_dart_list(contracted);
}

static tab_int init_father_priv(pyramid self,
                                tab_int tab)
{
  tab_int father;
  vertex mu;

  if(tab != NULL_TAB_INT)
  {
    father=tab;
    tab_int_reinitialize(father);
  }
  else
    father=FAIRE_tab_int();

  dart_list_first(Vertex_set);
  do
  {
    mu=dart_list_give_current(Vertex_set);
    tab_int_put_value(father,mu,mu);
  }
  while(dart_list_next(Vertex_set));

  return father;
}

static void update_vertex_set_priv(pyramid self,
                                   tab_int father,
                                   PFmerge merge)
{
   int edge;
   dart b;
   vertex mu,mu_b,f_mu;
   dart_list vertex_set;

   level_set_first(Levels,level_set_give_max_level(Levels));
   do
   {
     edge=level_set_current(Levels);	  
     for(b=-edge;b<=edge;b+=2*edge)
     {
       mu_b=tab_dart_give_value(Mu,b);
       f_mu=find_compress_priv(father,mu_b);
       tab_dart_put_value(Mu,b,f_mu);
       tab_int_put_value(Mu_1,f_mu,b);
     }
   }
   while(level_set_next(Levels));

   vertex_set=FAIRE_dart_list();
   /* Update of our vertex set => removal of contracted vertices */
   /* We use the self_loops list to store the surviving vertices */
   dart_list_first(Vertex_set);
   do
   {
     mu=dart_list_give_current(Vertex_set);
     f_mu=find_compress_priv(father,mu);
     if(f_mu == mu)
       dart_list_insert(vertex_set,mu);
     else
     {
       Vertex_value->merge(Vertex_value->data,self,
			   level_set_give_max_level(Levels)-3,
			   f_mu,mu);
     }
   }
   while(dart_list_next(Vertex_set));

   DEFAIRE_dart_list(Vertex_set);  
   Vertex_set=vertex_set;
}

static  void init_pyramid_priv(pyramid self,int dimx,int dimy)
{
  int i,j;
  int nb_edge;
  dart b,up,left,right,down;


  Sigma=FAIRE_tab_dart();
  Sigma_base=FAIRE_tab_dart();
  Mu=FAIRE_tab_dart();
  Mu_1=FAIRE_tab_int();

  /*  Node_values=FAIRE_tab_gen();*/
  Vertex_set=FAIRE_dart_list();
  Index_or=FAIRE_tab_int();
  Dart_emb=FAIRE_tab_dart();

  Dimx=dimx;
  Dimy=dimy;

  nb_edge = dimx*(dimy+1)+(dimx+1)*dimy;
  Background=dimx*dimy+1;

  up=1;
  left= dimx*(dimy+1)+1;
  right=left+1;
  down=-(dimx+1);
  for(j=0;j<dimy;j++)
  {
  	      /*  left column */
    tab_dart_put_value(Sigma_base,up,-left);	      
    tab_dart_put_value(Sigma_base,-left,down);  	      
    tab_dart_put_value(Sigma_base,down,right);  	      
    tab_dart_put_value(Sigma_base,right,up);	      
    
    
    for(i=0;i<dimx-1;i++)
    {
      up++;down--;right++;
  
      tab_dart_put_value(Sigma_base,up,-right+1);	    
      tab_dart_put_value(Sigma_base,-right+1,down);	    
      tab_dart_put_value(Sigma_base,down,right);	    
      tab_dart_put_value(Sigma_base,right,up);  		  
    }
    up++;down--;
    left=right+1;
    right+=2;
  }

  /* Encoding of the background vertex */

  left=dimx*(dimy+1)+1;
  tab_dart_put_value(Sigma_base,left,-1);		
  
  for(b=1;b<dimx;b++)
    tab_dart_put_value(Sigma_base,-b,-b-1);
  
  right=left+dimx;
  tab_dart_put_value(Sigma_base,-dimx,-right);

  for(b=right;b<=nb_edge-dimx-1;b+=dimx+1)
    tab_dart_put_value(Sigma_base,-b,-b-dimx-1);		

  right=  dimx*(dimy+1);
  tab_dart_put_value(Sigma_base,-nb_edge,right);


  for(b=right;b>right-dimx+1;b--)
    tab_dart_put_value(Sigma_base,b,b-1);

  left=right-dimx+1;
  tab_dart_put_value(Sigma_base,left,nb_edge-dimx);
  
  for(b=nb_edge-dimx;b>dimx*(dimy+1)+1;b-=(dimx+1))
    tab_dart_put_value(Sigma_base,b,b-dimx-1);
}

static void init_vertex_set_priv(pyramid self)
{
  int mu;
  dart b;

  /* Encoding of the array Mu */
  for(mu=1;mu<=Dimx*Dimy;mu++)
  {
    dart_list_insert(Vertex_set,mu);
    b=mu;
    do
    {
      tab_dart_put_value(Mu,b,mu);
      b=tab_dart_give_value(Sigma_base,b);
    }
    while(b!=mu);
    tab_int_put_value(Mu_1,mu,b);
  }
  /* Encoding of the background vertex */
  dart_list_insert(Vertex_set,Background);
  mu=Background;
  b=-1;
  do
  {
    tab_dart_put_value(Mu,b,mu);
    b=tab_dart_give_value(Sigma_base,b);
  }
  while(b!=-1);
  tab_int_put_value(Mu_1,mu,-1);
}
static void load_update_sigma_priv(pyramid self,tab_int levels)
{
  int edge;
  dart b,sig_b;
  int level_max=level_set_give_max_level(Levels);

  level_set_first(Levels,level_max);
  do
  {
    edge=level_set_current(Levels);
    for(b=-edge;b<=edge;b+=2*edge)
    {
      sig_b=load_compute_sig_succ_priv(self,levels,b);
      tab_dart_put_value(Sigma,b,sig_b);
    }
  }
  while(level_set_next(Levels));
}
static void load_set_vertice_priv(pyramid self,tab_int levels)
{
  vertex v=2;
  dart b_first,b,b_background;
  int level;
  tab_mark_dart encountered=FAIRE_tab_mark_dart();
  int level_max=level_set_give_max_level(Levels);

  Background=1;
  dart_list_insert(Vertex_set,Background);
  b_first=b=-1;
  do
    {
      level=tab_int_give_value(levels,ABS(b));
      if(level==level_max)
	{
	  tab_mark_dart_mark_dart(encountered,b);
	  tab_dart_put_value(Mu,b,Background);
	  b_background=b;
	}
      b=tab_dart_give_value(Sigma_base,b);
    }
  while(b != b_first);
  tab_int_put_value(Mu_1,Background,b_background);

  level_set_first(Levels,level_max);
  do
    {      
      b_first=b=level_set_current(Levels);
      if(tab_mark_dart_is_marked(encountered,b))
	continue;
      do
	{
	  tab_mark_dart_mark_dart(encountered,b);
	  tab_dart_put_value(Mu,b,v);
	  b=tab_dart_give_value(Sigma,b);
	}
      while(b != b_first);
      tab_int_put_value(Mu_1,v,b_first);
      dart_list_insert(Vertex_set,v);
      v++;
    }
  while(level_set_next(Levels));    
}
static dart load_compute_sig_succ_priv(pyramid self,
				       tab_int levels,
				       dart b)
{
  int level,level_max;

  level_max =level_set_give_max_level(Levels);
  do
    {
      b=next_boundary_dart_priv(self,levels,level_max,b);
      level=tab_int_give_value(levels,ABS(b));
    }
  while(level < level_max);

  return b;
}
extern move_type give_move(pyramid self, dart d)
{
  dart abs_d=ABS(d);
  int horizontal=Dimx*(Dimy+1);

  if(abs_d <= horizontal)
  {
    if(d<0)
      return right;
    return left;
  }
  if(d>0) return up;
  return down;
}
static void update_orientation_priv(pyramid self,
				    dart b,
				    dart sig)
{
  vertex mu,mu_b;
  int or_b,or_sig;
  int or_junction;
  move_type move_mb,move_sig;

  mu=tab_dart_give_value(Mu,b);
  mu_b=tab_dart_give_value(Mu,-b);

  if(mu==mu_b)
    return;
  
  if(b>0)
    or_b=tab_int_give_value(Index_or,b);
  else
    or_b=-tab_int_give_value(Index_or,-b);
  if(sig>0)
    or_sig=tab_int_give_value(Index_or,sig);
  else
    or_sig=-tab_int_give_value(Index_or,-sig);

  move_mb=give_move(self,
			 tab_dart_give_value(Dart_emb,-b));
  move_sig=give_move(self,
			  tab_dart_give_value(Dart_emb,sig));
  
  or_junction=move_type_compute_step_clock(REVERSE_SENSE(move_mb),
					   move_sig);
  or_b=or_b+or_sig+or_junction;
  if(b>0)
    tab_int_put_value(Index_or,b,or_b);
  else
    tab_int_put_value(Index_or,-b,-or_b);
  
  tab_dart_put_value(Dart_emb,-b,
		     tab_dart_give_value(Dart_emb,-sig));
}

static   dart_list get_open_parenthesis_priv(pyramid self,
					     vertex v,
					     tab_mark_dart open_mark)
{
  tab_int stack_dart=FAIRE_tab_int();
  tab_int stack_or=FAIRE_tab_int();
  int stack_cpt=-1;
  dart b_prec,b_suiv,b,b_first,top;
  vertex mu;
  int or,or_f,top_or,or_conn;
  move_type move,last_move,move_suiv,move_prec;
  dart_list open_parenthesis=FAIRE_dart_list();

  b=b_first=tab_int_give_value(Mu_1,v);
  or_f=0;
  b_prec=-sigma_1_priv(self,b);
  last_move=give_move(self,
			   tab_dart_give_value(Dart_emb,b_prec));
  last_move=REVERSE_SENSE(last_move);
  
  do
    {
      move=give_move(self,
			  tab_dart_give_value(Dart_emb,b));
      or_conn=move_type_compute_step_clock(last_move,move);
      or_f+=or_conn;

      if(b>0)
	or_f+=tab_int_give_value(Index_or,b);
      else
	or_f-=tab_int_give_value(Index_or,-b);
      last_move=REVERSE_SENSE(give_move(self,
					tab_dart_give_value(Dart_emb,
								 -b)));


      mu=tab_dart_give_value(Mu,-b);
      if(mu == v)
	{
	  top=0;
	  if(stack_cpt != -1)
	    {
	      top=tab_int_give_value(stack_dart,stack_cpt);
	      top_or=tab_int_give_value(stack_or,stack_cpt);
	    }
	  if(top==-b)
	    {
	      b_prec=-sigma_1_priv(self,b);	/* phi^{-1}(b) */
	      b_suiv=tab_dart_give_value(Sigma,-b); /* phi(b) */
	      if(b_prec == b_suiv)
		{
		  b_prec=-b_prec; /*sigma^{-1}(b)*/
		  if(tab_mark_dart_is_marked(open_mark,b_prec))
		    {
		      fprintf(stderr,"%d: open parenthesis\n",b);
		      tab_mark_dart_mark_dart(open_mark,b);
		      dart_list_insert(open_parenthesis,b);
		    }
		  else
		    {
		      fprintf(stderr,"%d: open parenthesis\n",-b);
		      tab_mark_dart_mark_dart(open_mark,-b);
		      dart_list_insert(open_parenthesis,-b);
		    }
		  
		}
	      else
		{
		  or=or_f-top_or-or_conn;
		  move=give_move(self,
				      tab_dart_give_value(Dart_emb,b));
		  move_prec=give_move(self,
					   tab_dart_give_value(Dart_emb,b_prec));
		  move_suiv=give_move(self,
					   tab_dart_give_value(Dart_emb,b_suiv));
		  move_prec=REVERSE_SENSE(move_prec);
		  move=REVERSE_SENSE(move);
		  or+=move_type_compute_step_clock(move_prec,move_suiv);
		  or-=move_type_compute_step_clock(move,move_suiv);
		  fprintf(stderr,"or(%d)=%d\n",b,or);
		
		  assert((or==4) || (or==-4));

		  if(or==4)
		    {
		  		  /*b is a closed parenthesis */
		      tab_mark_dart_mark_dart(open_mark,top);
		      dart_list_insert(open_parenthesis,top);
		    }
		  else
		    {
		      /* b is an open parenthesis */
		      tab_mark_dart_mark_dart(open_mark,-top);
		      dart_list_insert(open_parenthesis,-top);
		    }
		}
		  stack_cpt--;	      
	    }
	  else
	    {
	      stack_cpt++;
	      tab_int_put_value(stack_dart,stack_cpt,b);
	      tab_int_put_value(stack_or,stack_cpt,or_f);
	    }
	}
      b=tab_dart_give_value(Sigma,b);
    }
  while(b!=b_first);
  return open_parenthesis;
}  
  
static tab_mark_dart get_inclusions_priv(pyramid self,
					 vertex v,
					 dart_list included,
					 tab_mark_dart included_mark
					 )
{
  dart b,close;
  dart_list open_parenthesis;
  vertex mu;
  tab_mark_dart open_mark;

  open_mark    =FAIRE_tab_mark_dart();
  
  open_parenthesis=get_open_parenthesis_priv(self,v,open_mark);

  if(open_parenthesis == NULL_DART_LIST)
  {
    return open_mark;
  }
  if(dart_list_is_empty(open_parenthesis))
    {
      DEFAIRE_dart_list(open_parenthesis);
      return open_mark;
    }
  
  dart_list_first(open_parenthesis);  
  do
    {
      b=dart_list_give_current(open_parenthesis);
      close=-b;
      b=tab_dart_give_value(Sigma,b);
      while(b!= close)
	{		  
	  mu=tab_dart_give_value(Mu,-b);
	  if(mu !=v)
	    {	 /* insert if not already done */
	      if(!tab_mark_dart_is_marked(included_mark,mu))
	      {
		tab_mark_dart_mark_dart(included_mark,mu);
		if(included != NULL_DART_LIST)
		  dart_list_insert(included,mu);
	      }
	      b=tab_dart_give_value(Sigma,b);
	    }
	  else
	    if(tab_mark_dart_is_marked(open_mark,b))
	      b=-b;   /* step the self-loop */
	    else
	      b=tab_dart_give_value(Sigma,b);
	}
    }
  while(dart_list_next(open_parenthesis));
  DEFAIRE_dart_list(open_parenthesis);

  return open_mark;
}

					 
