/* tab_float.c (Combinatorial Pyramids project)
 *  Copyright (C) 2003 Luc Brun
 *  email: luc.brun@univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <assert.h>
#include "tab_float.h"
#include "objet.h"
#include "garbage.h"

#define Nb_line      SELF(nb_line)
#define Element_tab  SELF(element_tab)
#define Nb_pointer   SELF(nb_pointer)

#define ALLOC_LINE 200
#define ALLOC_NB_LINE 50

#define NUM_LIGNE(n) ((n) / ALLOC_LINE)
#define NUM_COLUMN(n)((n) % ALLOC_LINE)

static garbage line_collector    = NULL_GARBAGE;
static garbage tab_float_collector = NULL_GARBAGE;

		/* Fonctions liees au garbage */
static void    init_line_priv(float*);
static float*    alloc_line_priv();
static void    dealloc_line_priv(float *);
static tab_float alloc_tab_float_priv();
static void    dealloc_tab_float_priv(tab_float);
static void    realloc_tab_float_priv(tab_float);
static void    free_tab_float_priv(void*);

struct tab_float
{
  float** element_tab;
  int nb_line;
  int nb_pointer;
};

void ALLOC_tab_float()
{
  if(line_collector == NULL_GARBAGE)
    line_collector = FAIRE_garbage();
  if(tab_float_collector == NULL_GARBAGE)
    tab_float_collector = FAIRE_garbage();
}
void DEALLOC_tab_float()
{
  if(line_collector != NULL_GARBAGE)
    {
      DEFAIRE_garbage(line_collector,free);
      line_collector = NULL_GARBAGE;
    }
  if(tab_float_collector != NULL_GARBAGE)
    {
      DEFAIRE_garbage(tab_float_collector,free_tab_float_priv);
      tab_float_collector = NULL_GARBAGE;
    }
}
tab_float FAIRE_tab_float()
{
  tab_float self;

  if(line_collector == NULL_GARBAGE ||
     tab_float_collector != NULL_GARBAGE)
    ALLOC_tab_float();
     
  self = alloc_tab_float_priv();

  *Element_tab= alloc_line_priv();
  
  Nb_line = 1;

  return self;
}

/*---------------------------------------------------------------------------*/

void DEFAIRE_tab_float(tab_float self)
{  
  int i;

  assert(self);

  for(i=0;i<Nb_line;i++)
    dealloc_line_priv(Element_tab[i]);
  
  dealloc_tab_float_priv(self);
}

void tab_float_reinitialize(tab_float self)
{  
  int i;

  assert(self);

  for(i=1;i<Nb_line;i++)
    dealloc_line_priv(Element_tab[i]);

  init_line_priv(Element_tab[0]);
  Nb_line = 1;
}

/*---------------------------------------------------------------------------*/
/* renvoie la valeur contenue a l'indice ind */
/*---------------------------------------------------------------------------*/
float tab_float_give_value(tab_float self,int ind)
{
  int ligne,column;

  assert(self);

  ligne  = NUM_LIGNE(ind);
  column = NUM_COLUMN(ind);
  
/* ligne non encore allouee */
  if (ligne >= Nb_line) 
    return 0;
 
  return Element_tab[ligne][column];
}

/*---------------------------------------------------------------------------*/
/* met value a l'indice ind du tableau virtuel */
/*---------------------------------------------------------------------------*/
void tab_float_put_value(tab_float self,int ind,float value)
{
  int ligne,column;

  assert(self);

  ligne =  NUM_LIGNE(ind);
  column = NUM_COLUMN(ind);

  /* on ajoute autant de lignes que necessaire */
  while(ligne >= Nb_line)
    realloc_tab_float_priv(self);
  
  Element_tab[ligne][column] = value;
}
      
   /********************************************************************/
  /* Alloue une ligne si on a pas enclenche le garbage ou si celui ci */
 /* est nul                                                          */
/********************************************************************/
static float* alloc_line_priv()
{
  float* line;

  if(line_collector == NULL_GARBAGE)
    {
      line = (float*) calloc(ALLOC_LINE,sizeof(float));
      assert(line);
      return line;
    }
  if(garbage_is_empty(line_collector))
    {
      line = (float*) calloc(ALLOC_LINE,sizeof(float));
      assert(line);
      return line;
    }
  
  line =(float*) garbage_give_object(line_collector);
  assert(line);
  init_line_priv(line);

  return line;
}
   /*********************************************************************/
  /* Alloue un tab_float si on a pas enclenche le garbage ou si celui ci */
 /* est nul                                                           */
/*********************************************************************/
static tab_float alloc_tab_float_priv()
{
  tab_float self;
  
  if(tab_float_collector == NULL_GARBAGE)
    {
      self = (tab_float)malloc(sizeof(struct tab_float));
      Element_tab    = (float **) malloc(ALLOC_NB_LINE*sizeof(float*));
      Nb_pointer  = ALLOC_NB_LINE;
      
      return self;
    }
  if(garbage_is_empty(tab_float_collector))
    {
      self = (tab_float)malloc(sizeof(struct tab_float));
      Element_tab    = (float **) malloc(ALLOC_NB_LINE*sizeof(float*));
      Nb_pointer  = ALLOC_NB_LINE;

      return self;
    }

  self = garbage_give_object(tab_float_collector);

  return self;
}
  /****************************************************/
 /* Libere ou insere ds le garbage une ligne liberee */
/****************************************************/
static void dealloc_line_priv(float * line)
{
  assert(line);
  if(line_collector == NULL_GARBAGE)
    {
      free(line);
      return;
    }  
  garbage_insert_object(line_collector,line);
}
  /****************************************************/
 /* Libere ou insere ds le garbage un tab_float libere */
/****************************************************/
static void dealloc_tab_float_priv(tab_float self)
{
  assert(self);
  if(tab_float_collector == NULL_GARBAGE)
    {
      free(self);
      return;
    }
  garbage_insert_object(tab_float_collector,self);
}
  /****************************************************/
 /* Alloue dynamiquement de la place pour le tab_float */
/****************************************************/
static void realloc_tab_float_priv(tab_float self)
{
  assert(self);

  /* On regarde si on na pas depasse le nb max de LINE alloue */
  if(Nb_line == Nb_pointer)
    {
      Element_tab = (float**)realloc(Element_tab,
				   (Nb_pointer+ALLOC_NB_LINE)*
				   sizeof(float*));
      assert(Element_tab);
      Nb_pointer += ALLOC_NB_LINE;
    }
  Element_tab[Nb_line] = alloc_line_priv();
  Nb_line++;
}
   /*********************************************************************/
  /*  Fonction utilisee par le garbage pour liberer les tab_float encore */
 /*  en memoire.                                                      */
/*********************************************************************/
static void free_tab_float_priv(void* obj)
{
  tab_float self = obj;
  
  free(Element_tab);
  free(self);
}
static void init_line_priv(float* line)
{
  int i;
  float *ptr;
  
  ptr = line;
  for(i=0;i<ALLOC_LINE;i++)
    {
      *ptr = 0;
      ptr++;
    }
}




